#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
测试 embedding loader

Authors: panxu(panxu@baidu.com)
Date:    2020/06/25 09:56:00
"""
import os
from easytext.tests import ROOT_PATH
from easytext.tests import ASSERT

from easytext.data import GloveLoader


def test_glove_loader():
    pretrained_file_path = "data/easytext/tests/pretrained/word_embedding_sample.3d.txt"
    pretrained_file_path = os.path.join(ROOT_PATH, pretrained_file_path)

    glove_loader = GloveLoader(embedding_dim=3,
                               pretrained_file_path=pretrained_file_path)

    embedding_dict = glove_loader.load()
    expect_embedding_dict = {"a": [1.0, 2.0, 3.0],
                             "b": [4.0, 5.0, 6.0],
                             "美丽": [7.0, 8.0, 9.0]}

    ASSERT.assertDictEqual(expect_embedding_dict, embedding_dict)
    ASSERT.assertEqual(glove_loader.embedding_dim, 3)



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
brief

Authors: PanXu
Date:    2020/10/16 16:40:00
"""
import os
import pytest

from torch.utils.tensorboard import SummaryWriter

from easytext.tests import ROOT_PATH


@pytest.fixture(scope="package")
def summary_writer():
    log_dir = "data/tensorboard"
    log_dir = os.path.join(ROOT_PATH, log_dir)

    return SummaryWriter(log_dir=log_dir)



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
"""
brief

Authors: panxu(panxu@baidu.com)
Date:    2020/06/01 15:40:00
"""
import logging
import math
from typing import Optional, List, Tuple

import torch
from torch import LongTensor, BoolTensor


def has_tensor(obj) -> bool:
    """
    检查 对象中是否包含 Tensor
    :param obj:
    :return: True: 包含 Tensor; False: 没有 Tensor
    """
    if isinstance(obj, torch.Tensor):
        return True
    elif isinstance(obj, dict):
        return any(has_tensor(value) for value in obj.values())
    elif isinstance(obj, (list, tuple)):
        return any(has_tensor(item) for item in obj)
    else:
        return False


def sequence_mask(sequence: LongTensor, padding_index: int = 0) -> BoolTensor:
    """
    计算 sequence 序列的 mask
    :param sequence: sequence index, 维度是 (B, SeqLen)
    :param padding_index: padding 的index, 一般是 0，也可以根据自己的padding index 来设置。
    :return: sequence mask, 注意是 BoolTensor, 根据需要可以转化成 Float 或者 Long
    """

    if sequence.dim() != 2:
        raise RuntimeError(f"Sequence 的维度 {sequence.dim()} 不是 2，也就是 (B, SeqLen)")

    return sequence != padding_index


def logsumexp(tensor: torch.Tensor,
              dim: int = -1,
              keepdim: bool = False) -> torch.Tensor:
    """
    `tensor.exp().sum(dim, keep=keepdim).log()`。 就是: log(exp(S01 + S02 + ...) + exp(S11 + S12 + ...))
    也被叫做 logadd 是 条件随机场中计算边缘概率用的公式。
    :param tensor: 要计算的 tensor
    :param dim: 沿着那个维度来进行 sum
    :param keepdim: 是否保持原先的维度, 默认是 False
    :return: logadd 值
    """

    # 计算 max_score, 是为了避免 tensor 中的数值太小或者太大而导致的溢出
    max_score, _ = tensor.max(dim, keepdim=keepdim)
    if keepdim:
        stable_vec = tensor - max_score
    else:
        stable_vec = tensor - max_score.unsqueeze(dim)
    return max_score + (stable_vec.exp().sum(dim, keepdim=keepdim)).log()


def viterbi_decode(tag_sequence: torch.Tensor,
                   transition_matrix: torch.Tensor,
                   tag_observations: Optional[List[int]] = None,
                   allowed_start_transitions: torch.Tensor = None,
                   allowed_end_transitions: torch.Tensor = None) -> Tuple[List[int], torch.Tensor]:
    """
    viterbi 解码
    :param tag_sequence: tag sequence, shape (sequence_len, num_tag).
    :param transition_matrix: 转移矩阵, shape (num_tag, num_tag)
    :param tag_observations: 一般情况下这个参数是不会被设置的。A list of length ``sequence_length`` containing the class ids of observed
    elements in the sequence, with unobserved elements being set to -1. Note that
    it is possible to provide evidence which results in degenerate labelings if
    the sequences of tags you provide as evidence cannot transition between each
    other, or those transitions are extremely unlikely. In this situation we log a
    warning, but the responsibility for providing self-consistent evidence ultimately
    lies with the user.
    :param allowed_start_transitions: <BOS>, shape (num_tags,)，有额外增加 START 的时候对 转移矩阵的额外设置。
    :param allowed_end_transitions: <EOS>, shape (num_tag,)，有额外增加 STOP 的时候对转移矩阵的额外设置。
    :return: 1. viterbi_path : List[int] - viterbi 路径，也就是 max socre 路径。
             2. viterbi_score : torch.Tensor - viterbi 路径的分数, 就只有一个值。
    """

    sequence_length, num_tags = list(tag_sequence.size())

    has_start_end_restrictions = allowed_end_transitions is not None or allowed_start_transitions is not None

    if has_start_end_restrictions:

        if allowed_end_transitions is None:
            allowed_end_transitions = torch.zeros(num_tags)
        if allowed_start_transitions is None:
            allowed_start_transitions = torch.zeros(num_tags)

        num_tags = num_tags + 2
        new_transition_matrix = torch.zeros(num_tags, num_tags)
        new_transition_matrix[:-2, :-2] = transition_matrix

        # Start and end transitions are fully defined, but cannot transition between each other.
        # pylint: disable=not-callable
        allowed_start_transitions = torch.cat([allowed_start_transitions, torch.tensor([-math.inf, -math.inf])])
        allowed_end_transitions = torch.cat([allowed_end_transitions, torch.tensor([-math.inf, -math.inf])])
        # pylint: enable=not-callable

        # First define how we may transition FROM the start and end tags.
        new_transition_matrix[-2, :] = allowed_start_transitions
        # We cannot transition from the end tag to any tag.
        new_transition_matrix[-1, :] = -math.inf

        new_transition_matrix[:, -1] = allowed_end_transitions
        # We cannot transition to the start tag from any tag.
        new_transition_matrix[:, -2] = -math.inf

        transition_matrix = new_transition_matrix

    if tag_observations:
        if len(tag_observations) != sequence_length:
            raise RuntimeError("Observations were provided, but they were not the same length "
                                     "as the sequence. Found sequence of length: {} and evidence: {}"
                                     .format(sequence_length, tag_observations))
    else:
        tag_observations = [-1 for _ in range(sequence_length)]


    if has_start_end_restrictions:
        tag_observations = [num_tags - 2] + tag_observations + [num_tags - 1]
        zero_sentinel = torch.zeros(1, num_tags)
        extra_tags_sentinel = torch.ones(sequence_length, 2) * -math.inf
        tag_sequence = torch.cat([tag_sequence, extra_tags_sentinel], -1)
        tag_sequence = torch.cat([zero_sentinel, tag_sequence, zero_sentinel], 0)
        sequence_length = tag_sequence.size(0)

    path_scores = []
    path_indices = []

    if tag_observations[0] != -1:
        one_hot = torch.zeros(num_tags)
        one_hot[tag_observations[0]] = 100000.
        path_scores.append(one_hot)
    else:
        path_scores.append(tag_sequence[0, :])

    # Evaluate the scores for all possible paths.
    for timestep in range(1, sequence_length):
        # Add pairwise potentials to current scores.
        summed_potentials = path_scores[timestep - 1].unsqueeze(-1) + transition_matrix
        scores, paths = torch.max(summed_potentials, 0)

        # If we have an observation for this timestep, use it
        # instead of the distribution over tags.
        observation = tag_observations[timestep]
        # Warn the user if they have passed
        # invalid/extremely unlikely evidence.
        if tag_observations[timestep - 1] != -1 and observation != -1:
            if transition_matrix[tag_observations[timestep - 1], observation] < -10000:
                logging.warning("The pairwise potential between tags you have passed as "
                               "observations is extremely unlikely. Double check your evidence "
                               "or transition potentials!")
        if observation != -1:
            one_hot = torch.zeros(num_tags)
            one_hot[observation] = 100000.
            path_scores.append(one_hot)
        else:
            path_scores.append(tag_sequence[timestep, :] + scores.squeeze())
        path_indices.append(paths.squeeze())

    # Construct the most likely sequence backwards.
    viterbi_score, best_path = torch.max(path_scores[-1], 0)
    viterbi_path = [int(best_path.numpy())]
    for backward_timestep in reversed(path_indices):
        viterbi_path.append(int(backward_timestep[viterbi_path[-1]]))
    # Reverse the backward path.
    viterbi_path.reverse()

    if has_start_end_restrictions:
        viterbi_path = viterbi_path[1:-1]
    return viterbi_path, viterbi_score


def masked_softmax(vector: torch.FloatTensor,
                   mask: torch.ByteTensor):
    """
    计算带有 masked 的 softmax
    :param vector: shape: (B, seq_len)
    :param mask: shape: (B, seq_len),
    :return: (B, seq_len)
    """
    exp_vector = vector.exp()

    masked_vector = exp_vector * mask.float()

    return masked_vector / torch.sum(masked_vector, dim=-1, keepdim=True)






#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
测试 f1 metric

Authors: PanXu
Date:    2020/12/21 07:24:00
"""
from typing import Dict, Tuple, Union, List

import torch
from torch import Tensor
from torch.distributed import ReduceOp

from easytext.metrics import F1Metric

from easytext.tests import ASSERT


class _DemoF1Metric(F1Metric):

    def __init__(self):
        labels = ["a", "b", "c"]
        super().__init__(labels=labels)

        for i, label in enumerate(labels):

            self._true_positives[label] = i + 1
            self._false_negatives[label] = i + 2
            self._false_positives[label] = i + 3

    @property
    def true_positives(self):
        return self._true_positives

    @property
    def false_negatives(self):
        return self._false_negatives

    @property
    def false_positives(self):
        return self._false_positives

    def __call__(self, prediction_labels: torch.Tensor, gold_labels: torch.Tensor, mask: torch.LongTensor) -> Dict:
        pass


def test_synchronized_data():
    """
    测试 to_synchronized_data 和 from_synchronized_data
    :return:
    """

    demo_metric = _DemoF1Metric()

    sync_data, op = demo_metric.to_synchronized_data()

    true_positives = sync_data["true_positives"]
    false_positives = sync_data["false_positives"]
    false_negatives = sync_data["false_negatives"]

    expect_values = [v for _, v in demo_metric.true_positives.items()]
    ASSERT.assertListEqual(expect_values, true_positives.tolist())

    expect_values = [v for _, v in demo_metric.false_positives.items()]
    ASSERT.assertListEqual(expect_values, false_positives.tolist())

    expect_values = [v for _, v in demo_metric._false_negatives.items()]
    ASSERT.assertListEqual(expect_values, false_negatives.tolist())

    expect_true_positives = dict(demo_metric.true_positives)
    expect_false_positives = dict(demo_metric.false_positives)
    expect_false_negatives = dict(demo_metric.false_negatives)

    demo_metric.from_synchronized_data(sync_data=sync_data, reduce_op=op)

    ASSERT.assertDictEqual(expect_true_positives, demo_metric.true_positives)
    ASSERT.assertDictEqual(expect_false_positives, demo_metric.false_positives)
    ASSERT.assertDictEqual(expect_false_negatives, demo_metric.false_negatives)






#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
分布式训练用到的相关工具

Authors: PanXu
Date:    2020/11/09 18:08:00
"""
from typing import Union, Optional
from torch import distributed


class NotCall:
    pass


class DistributedFuncWrapper:
    """
    分布式训练函数包装器, 会根据指定的进程 rank 进行处理
    """

    def __init__(self, dst_rank: Optional[int] = 0):
        """
        初始化
        :param dst_rank: 需要调用函数的 dst rank, 如果是 None, 则表示直接返回不会进行分布式处理
        """
        self._dst_rank = dst_rank

    @property
    def dst_rank(self):
        return self._dst_rank

    def __call__(self, func, *args, **kwargs) -> Union[NotCall, object]:
        """
        在指定的 dst rank 进程运行该函数
        :param func: 函数对象
        :param args: 函数的参数
        :param kwargs: 函数的参数
        :return: 如果没有被运行, 返回 NotCall 对象; 否则, 返回 函数的返回值。
        """

        if self._dst_rank is not None:
            if distributed.get_rank() == self._dst_rank:
                return func(*args, **kwargs)
            return NotCall()
        else:
            return func(*args, **kwargs)



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
"""
brief

Authors: panxu(panxu@baidu.com)
Date:    2020/06/01 16:10:00
"""

from .optimizer_factory import OptimizerFactory
from .lr_scheduler_factory import LRSchedulerFactory



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
brief

Authors: PanXu
Date:    2020/08/21 09:34:00
"""

from .grad_rescaled import GradRescaled
from .grad_clip_value import GradClipValue
from .grad_normalize import GradNormalize



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
通用的 预训练词向量载入器

Authors: panxu(panxu@baidu.com)
Date:    2020/06/24 11:44:00
"""
import logging
from tqdm import tqdm
from typing import Dict, List

from easytext.data.pretrained_word_embedding_loader import PretrainedWordEmbeddingLoader
from easytext.component.register import EasytextRegister


@EasytextRegister.register()
class GeneralPretrainedWordEmbeddingLoader(PretrainedWordEmbeddingLoader):
    """
    一般的通用预训练好的词向量载入器。试用范围: 词向量是一个文本文件，每一行是一个词向量。格式是:

    [word] [value] [value] ...

    比如 Glove6B 这些
    """

    def __init__(self,
                 embedding_dim: int,
                 pretrained_file_path: str,
                 encoding="utf-8",
                 separator: str = " ",
                 skip_num_line: int = 0,
                 max_size: int = None):
        """
        初始化,
        :param embedding_dim: embedding dim 大小
        :param pretrained_file_path: 预训练好的词向量文件路径
        :param encoding: 文件编码格式，默认是 "utf-8"
        :param separator: 每一行词向量的分隔符，默认是 " " 空格。
        :param skip_num_line: 跳过的行数，因为有些词向量头部 不是词向量。
        :param max_size: 载入的最大向量数量。因为某些词向量可能很大，比如 glove.840B.300d.txt 有 5G 大小，
        通过该参数进行限制，在比如测试阶段是非常有益的，或者一些内存小的机器上。
        """
        self._embedding_dim = embedding_dim
        self._pretrained_file_path = pretrained_file_path
        self._separator = separator
        self._encoding = encoding
        self._skip_num_line = skip_num_line
        self._max_size = max_size
        self._embedding_dict = None

    def load(self) -> Dict[str, List[float]]:
        """
        预训练的载入方法
        return: embedding dict, key 是 token, value: List[float] 向量，例如:
        {"the": [1.0, 2.0, ...]}
        """
        if self._embedding_dict is not None:
            logging.warning(f"embedding_table 已经由 {self._pretrained_file_path} 通过 load 构建过了, "
                            f"现在正在再次执行 load, 先前构建的会被新的取代.")

        self._embedding_dict = dict()
        logging.info(f"Begin load pretrained embedding from {self._pretrained_file_path} ...")

        with open(self._pretrained_file_path, encoding=self._encoding) as f:
            for line_no, line in tqdm(enumerate(f)):

                if line_no < self._skip_num_line:
                    continue

                line = line.rstrip()
                if len(line) == 0:  # 去掉空行
                    continue

                if self._max_size is not None:
                    if (line_no - self._skip_num_line) >= self._max_size:
                        break

                items = line.split(sep=self._separator)
                token = items[0]
                vec = [float(v) for v in items[1:]]

                assert len(vec) == self._embedding_dim, f"读取的向量维度: {len(vec)} != 设置的维度: {self._embedding_dim}"

                self._embedding_dict[token] = vec
        logging.info(f"End load pretrained embedding from {self._pretrained_file_path}.")
        return self._embedding_dict

    @property
    def embedding_dim(self) -> int:
        """
        词向量维度
        return: 词向量维度
        """
        return self._embedding_dim

    @property
    def embedding_dict(self) -> Dict[str, List[float]]:

        if self._embedding_dict is None:
            logging.warning(f"embedding_table 没有构建，请先调用 load")
        return self._embedding_dict




#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
预训练好的词汇表，带有预训练好的词向量

Authors: panxu(panxu@baidu.com)
Date:    2020/06/23 11:49:00
"""
import os
from typing import Iterable, List, Dict, Union
import torch
from .vocabulary import IVocabulary, Vocabulary

from easytext.data.pretrained_word_embedding_loader import PretrainedWordEmbeddingLoader


class PretrainedVocabulary(IVocabulary):
    """
    带有预训练词向量的词汇表。
    """

    EMBEDDING_MATRIX_FILE_NAME = "embedding_matrix.pt"

    def __init__(self,
                 vocabulary: Vocabulary,
                 pretrained_word_embedding_loader: PretrainedWordEmbeddingLoader,
                 ):
        """
        初始化
        :param vocabulary: 词汇表
        :param pretrained_word_embedding_loader: word embedding 载入器
        """

        self._vocabulary = vocabulary

        if pretrained_word_embedding_loader is not None:

            if pretrained_word_embedding_loader.embedding_dict is None:
                pretrained_word_embedding_loader.load()

            embedding_dict = pretrained_word_embedding_loader.embedding_dict

            embeddings = list()

            for index in range(self._vocabulary.size):
                token = self._vocabulary.token(index)

                if token in embedding_dict:
                    embeddings.append(embedding_dict[token])
                else:
                    empty_vec = [0.] * pretrained_word_embedding_loader.embedding_dim
                    embeddings.append(empty_vec)

            self._embedding_matrix = torch.tensor(embeddings, dtype=torch.float)
        else:
            self._embedding_matrix = None

    @property
    def embedding_matrix(self) -> torch.Tensor:
        """
        词向量 matrix
        """
        return self._embedding_matrix

    def save_to_file(self, directory: str) -> "PretrainedVocabulary":
        self._vocabulary.save_to_file(directory)

        # 将 embedding matrix 存起来
        embedding_matrix_file_path = os.path.join(directory, PretrainedVocabulary.EMBEDDING_MATRIX_FILE_NAME)
        torch.save(self._embedding_matrix, embedding_matrix_file_path)

        return self

    @classmethod
    def from_file(cls, directory: str) -> "PretrainedVocabulary":
        vocabulary = Vocabulary.from_file(directory)

        pretrianed_vocabulary = cls(pretrained_word_embedding_loader=None,
                                    vocabulary=vocabulary)
        embedding_matrix_file_path = os.path.join(directory, PretrainedVocabulary.EMBEDDING_MATRIX_FILE_NAME)
        pretrianed_vocabulary._embedding_matrix = torch.load(embedding_matrix_file_path)

        return pretrianed_vocabulary

    def __len__(self):
        return len(self._vocabulary)

    @property
    def unk(self):
        return self._vocabulary.unk

    @property
    def padding(self):
        return self._vocabulary.padding

    @property
    def other_special_tokens(self):
        return self._vocabulary.other_special_tokens

    @property
    def padding_index(self) -> Union[None, int]:
        """
        :return: 获取 padding 的 index, 如果 padding 没有设置，那么返回 None; 否则，返回实际的index.
        """
        return self._vocabulary.padding_index

    def index(self, token: str) -> int:
        """
        获取token的index
        :param token: 输入的token
        :return: token 的 index
        """

        return self._vocabulary.index(token)

    def token(self, index: int) -> str:
        """
        获取 index 的 token
        :param index: 指定 index
        :return: 当前 index 的 token
        """
        return self._vocabulary.token(index)

    @property
    def size(self):
        return self._vocabulary.size




#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
测试 attention seq2vec

Authors: PanXu
Date:    2020/07/18 16:06:00
"""
import pytest

import torch
from torch.nn import Parameter
from torch import FloatTensor, LongTensor, ByteTensor

from easytext.modules.seq2vec import AttentionSeq2Vec

from easytext.tests import ASSERT


@pytest.fixture(scope="package")
def inputs():
    sequence = torch.FloatTensor(
        [
            [[1., 2.], [3., 4.], [5., 6.], [9., 8.]],
            [[7., 1.], [2., 8.], [9., 3.], [5., 2.]]
        ]
    )
    mask = torch.ByteTensor(
        [
            [1, 1, 1, 0],
            [1, 1, 0, 0]
        ]
    )
    return sequence, mask


def test_attention_seq2vec_with_mask(inputs):
    """
    测试 attention seq2vec
    :return:
    """

    sequence, mask = inputs

    encoder = AttentionSeq2Vec(input_size=2,
                               query_hidden_size=3,
                               value_hidden_size=None)

    encoder.wk.weight = Parameter(FloatTensor([
        [0.1, 0.2],
        [0.3, 0.4],
        [0.5, 0.6]
    ]))
    encoder.wk.bias = Parameter(FloatTensor([0.2, 0.4, 0.6]))

    encoder.attention.weight = Parameter(FloatTensor(
        [
            [0.6, 0.2, 7]
        ]
    ))

    vec = encoder(sequence=sequence, mask=mask)

    ASSERT.assertEqual((2, 2), vec.size())

    expect = torch.tensor([[3.23450, 4.23450],
                           [4.37470, 4.67550]])

    vec1d = vec.view(-1).tolist()
    expect1d = expect.view(-1).tolist()

    for expect_data, vec_data in zip(expect1d, vec1d):
        ASSERT.assertAlmostEqual(expect_data, vec_data, delta=1e-4)


def test_attention_seq2vec_no_mask(inputs):
    """
    测试 attention seq2vec
    :return:
    """

    sequence, mask = inputs

    encoder = AttentionSeq2Vec(input_size=2,
                               query_hidden_size=3,
                               value_hidden_size=None)

    encoder.wk.weight = Parameter(FloatTensor([
        [0.1, 0.2],
        [0.3, 0.4],
        [0.5, 0.6]
    ]))
    encoder.wk.bias = Parameter(FloatTensor([0.2, 0.4, 0.6]))

    encoder.attention.weight = Parameter(FloatTensor(
        [
            [0.6, 0.2, 7]
        ]
    ))

    vec = encoder(sequence=sequence, mask=None)

    print(vec)

    ASSERT.assertEqual((2, 2), vec.size())

    expect = torch.tensor([[4.8455, 5.2867],
                           [5.7232, 3.6037]])

    vec1d = vec.view(-1).tolist()
    expect1d = expect.view(-1).tolist()

    for expect_data, vec_data in zip(expect1d, vec1d):
        ASSERT.assertAlmostEqual(expect_data, vec_data, delta=1e-4)



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2021 PanXu, Inc. All Rights Reserved
#
"""
bio 的 schema 转换

Authors: PanXu
Date:    2021/02/03 09:39:00
"""

from typing import List


def ibo1_to_bio(sequence_label: List[str]) -> List[str]:
    """
    IBO1 的序列标签格式转换成 bio 格式。ibo1格式是指 "I-Label" 作为一个序列的开始, 直到遇到 O。
    "B-Label" 是指连续 "I-Label" 是多个 span 的时候，除了第一个span，其他都是 "B-Label"。
    另外 BIO 的格式也兼容，比如 "B-Lable", "I-Label" 这种BIO格式也是正确的。
    例子如下:
    "I-Label1 I-Label1 O I-Label1 I-Label1 I-Label2 I-Label2 O I-Label1 I-Label1 B-Label1 I-Label1 O B-Label I-Label"

    这里包含的spans:
    "[I-Label1 I-Label1] O [I-Label1 I-Label1] [I-Label2 I-Label2] O [I-Label1 I-Label1] [B-Label1 I-Label1] O [B-Label I-Label"]

    :param sequence_label:
    :return:
    """

    bio: List[str] = list()

    idel_state = 0
    span_state = 1
    state = idel_state

    for i, label in enumerate(sequence_label):

        if state == idel_state:
            if label == "O":
                state = idel_state
                bio.append(label)
            elif label[0] == "I":
                state = span_state
                # 将 I-label 转换成 B-label
                b_label = "B" + label[1:]
                bio.append(b_label)
            elif label[0] == "B":
                # 这种情况是 BIO 标注，认为是对的
                state = span_state
                bio.append(label)
        elif state == span_state:
            if label == "O":
                state = idel_state
                bio.append(label)
            elif label[0] == "I":
                if bio[-1][1:] == label[1:]:
                    # 与先前的相同，也就是 I-label1 I-label1 的情况，所以直接append即可
                    state = span_state
                    bio.append(label)
                else:
                    # 与先前不一样说明是新的，也就是 I-label1 I-label2 的情况
                    state = span_state
                    b_label = "B" + label[1:]
                    bio.append(b_label)
            elif label[0] == "B":
                # 这是新的了 也就是 I-label1 B-label1 情况
                state = span_state
                bio.append(label)
        else:
            raise RuntimeError(f"非法的状态: {state}")

    return bio


def bmes_to_bio(sequence_labels: List[str]) -> List[str]:
    """
    将 BMES schema 转换成 BIO schema
    :param sequence_labels: BMES schema 的 label 序列
    :return: BIO schema 的 label 序列
    """
    bio_sequence_labels = list()

    for label in sequence_labels:

        if label[0] == "B":
            bio_label = "B" + label[1:]
        elif label[0] == "M":
            bio_label = "I" + label[1:]
        elif label[0] == "E":
            bio_label = "I" + label[1:]
        elif label[0] == "S":
            bio_label = "B" + label[1:]
        elif label[0] == "O":
            bio_label = "O" + label[1:]
        else:
            raise RuntimeError(f"非法的 BMES label: {label}")

        bio_sequence_labels.append(bio_label)

    return bio_sequence_labels



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
Component Factory

Authors: PanXu
Date:    2020/10/27 16:14:00
"""
import traceback
import logging
import inspect
import copy
from collections import OrderedDict

from easytext.component import Component
from easytext.component.component_builtin_key import ComponentBuiltinKey
from easytext.component.register import Registry
from easytext.utils.json_util import json2str


class ComponentFactory:
    """
    Component Factory
    """

    def __init__(self, is_training: bool):
        """
        初始化
        :param is_training: True: training 状态创建 component; False: 非training状态创建 component
        """
        self._is_training = is_training
        self._registry = Registry()

    def _create_by_object(self, path: str, param_dict: OrderedDict):
        """
        通过 object 来得到 object, 因为 object 是之前创建好的，直接过去就好
        :param path:
        :param param_dict:
        :return:
        """
        object_path = param_dict[ComponentBuiltinKey.OBJECT]
        return self._registry.find_object(object_path)

    def _create_by_type(self, path: str, param_dict: OrderedDict):
        """
        通过 type 和 name space 创建 object
        :param path:
        :param param_dict:
        :return:
        """
        component_type = param_dict.pop(ComponentBuiltinKey.TYPENAME)
        name_space = param_dict.pop(ComponentBuiltinKey.NAME_SPACE)

        cls = self._registry.find_class(name=component_type, name_space=name_space)

        if cls is None:
            raise RuntimeError(f"{name_space}:{component_type} 没有被注册")

        for param_name, param_value in param_dict.items():

            if isinstance(param_value, OrderedDict):
                sub_path = f"{path}.{param_name}"
                v_obj = self._create(path=sub_path, param_dict=param_value)
                param_dict[param_name] = v_obj
            else:
                # 不用处理
                pass

        # 增加 is_training 参数
        need_is_training_parameter = False
        if inspect.isclass(cls):
            if issubclass(cls, Component):
                need_is_training_parameter = True
            else:
                # 非 Component 类，不做任何处理
                pass
        elif inspect.isfunction(cls):
            if ComponentBuiltinKey.IS_TRAINING in inspect.getfullargspec(cls).args:
                need_is_training_parameter = True

        else:
            raise RuntimeError(f"{cls} 错误! 应该是 函数 或者是 类的静态函数, 不能是类函数或者成员函数")

        if need_is_training_parameter and (ComponentBuiltinKey.IS_TRAINING not in param_dict):
            param_dict[ComponentBuiltinKey.IS_TRAINING] = self._is_training

        try:
            obj = cls(**param_dict)
        except TypeError as type_error:
            logging.fatal(f"Exception: {type_error} for {cls}")
            logging.fatal(traceback.format_exc())
            raise type_error

        self._registry.register_object(name=path, obj=obj)
        return obj

    def _create_by_raw_dict(self, path: str, param_dict: OrderedDict):
        """
        没有 type 和 name space, 该字典就是参数
        :param path:
        :param param_dict:
        :return:
        """

        for param_name, param_value in param_dict.items():

            if isinstance(param_value, OrderedDict):
                sub_path = f"{path}.{param_name}"
                v_obj = self._create(path=sub_path, param_dict=param_value)
                param_dict[param_name] = v_obj
            else:
                # 不用处理
                pass

        return param_dict

    def _create(self, path: str, param_dict: OrderedDict):

        param_dict: OrderedDict = param_dict

        if ComponentBuiltinKey.OBJECT in param_dict:
            return self._create_by_object(path=path, param_dict=param_dict)

        elif ComponentBuiltinKey.TYPENAME in param_dict and ComponentBuiltinKey.NAME_SPACE in param_dict:
            return self._create_by_type(path=path, param_dict=param_dict)

        elif ComponentBuiltinKey.TYPENAME in param_dict and ComponentBuiltinKey.NAME_SPACE not in param_dict:
            raise RuntimeError(f"构建 {path} 错误, "
                               f"{ComponentBuiltinKey.TYPENAME} 与 {ComponentBuiltinKey.NAME_SPACE} 必须同时出现")

        elif ComponentBuiltinKey.TYPENAME not in param_dict and ComponentBuiltinKey.NAME_SPACE in param_dict:
            raise RuntimeError(f"构建 {path} 错误, "
                               f"{ComponentBuiltinKey.TYPENAME} 与 {ComponentBuiltinKey.NAME_SPACE} 必须同时出现")

        else:
            # 这种情况是指，参数就是一个字典
            return self._create_by_raw_dict(path=path, param_dict=param_dict)

    def create(self, config: OrderedDict):
        """
        创建对象工厂
        :param config: config 字典, 是 OrderedDict, 其中的 key 会按照顺序执行
        :return:
        """

        assert isinstance(config, OrderedDict), f"param_dict type: {type(config)} 不是 OrderedDict"

        parsed_config = copy.deepcopy(config)

        for obj_name, param_dict in parsed_config.items():

            if isinstance(param_dict, OrderedDict):
                parsed_config[obj_name] = self._create(obj_name, param_dict=param_dict)
            else:
                # 对于非字典的根节点下的是基础类型，不用做处理
                pass
        return parsed_config



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
测试 max label index decoder

Authors: PanXu
Date:    2020/07/06 10:23:00
"""
import torch

from easytext.tests import ASSERT
from easytext.label_decoder import MaxLabelIndexDecoder


def test_max_label_index_decoder():
    """
    测试 max label index
    :return:
    """
    decoder = MaxLabelIndexDecoder()

    logits = torch.tensor([[0.1, 0.9], [0.3, 0.7], [0.8, 0.2]])

    label_indices = decoder(logits=logits)

    expect = [1, 1, 0]

    ASSERT.assertListEqual(expect, label_indices.tolist())



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
模型组件注册器

Authors: PanXu
Date:    2020/10/27 15:03:00
"""

from easytext.component import ComponentBuiltinKey
from easytext.component.register import ComponentRegister
from easytext.component.register.component_register import T


class EasytextRegister:
    """
    easytext 系统使用的注册器，外部永远不要使用这个注册器，而应该使用 ComponentRegister
    """

    @classmethod
    def register(cls, typename: str = None, is_allowed_exist: bool = False) -> T:
        return ComponentRegister.register(typename=typename,
                                          name_space=ComponentBuiltinKey.EASYTEXT_NAME_SPACE,
                                          is_allowed_exist=is_allowed_exist)





#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
组件

Authors: PanXu
Date:    2020/11/01 12:48:00
"""


class Component:
    """
    组件, 在配置文件中出现的对象可能会需要继承该类，该类在创建对象的时候是区分是否是 training 状态
    """

    def __init__(self, is_training: bool):
        """
        创建对象的初始化函数
        :param is_training: True: 表示训练状态; False: 表示非训练状态. 该参数通过 self._is_training 能够获取到
        """
        self._is_training = is_training





#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
sequence label decoder

Authors: PanXu
Date:    2020/07/05 11:27:00
"""
from typing import List

import torch

from easytext.data import LabelVocabulary
from easytext.utils import bio as BIO

from .label_decoder import LabelDecoder


class SequenceLabelDecoder(LabelDecoder):
    """
    对序列label解码成用户需要的
    """

    def __init__(self, label_vocabulary: LabelVocabulary):
        self._label_vocabulary = label_vocabulary

    def __call__(self, label_indices: torch.LongTensor, mask: torch.ByteTensor) -> List:
        """
        将 label index 解码 成span

        batch_sequence_label shape:(B, seq_len)  (B-T: 0, I-T: 1, O: 2)
        label index:
        [[0, 1, 2],
         [2, 0, 1]]

         对应label序列是:
         [[B, I, O],
          [O, B, I]]

         解码成:

         [[{"label": T, "begin": 0, "end": 2}],
          [{"label": T, "begin": 1, "end": 3}]]

        :param label_indices: label index
        :param mask: mask
        :return: span list
        """
        spans = BIO.decode_label_index_to_span(
            batch_sequence_label_index=label_indices,
            mask=mask,
            vocabulary=self._label_vocabulary)

        return spans



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
"""
训练器

Authors: panxu(panxu@baidu.com)
Date:    2020/05/16 00:34:00
"""

import os
import torch
import logging
from tqdm import tqdm
import shutil

from typing import Optional, Union, List

from torch.utils.data import DataLoader
from torch.utils.data.distributed import DistributedSampler
from torch.optim.lr_scheduler import ReduceLROnPlateau
from torch.nn.parallel import DistributedDataParallel
from torch import distributed as TorchDist
from torch.distributed import ReduceOp

from easytext.model import Model
from easytext.loss import Loss
from easytext.optimizer import OptimizerFactory
from easytext.optimizer import LRSchedulerFactory
from easytext.data.model_collate import ModelInputs
from easytext.metrics import ModelMetricAdapter
from easytext.utils.json_util import json2str
from easytext.utils.nn import cuda_util
from easytext.utils.distributed.distributed_util import DistributedFuncWrapper
from easytext.utils.distributed import Sync
from easytext.trainer.metric_tracker import MetricTracker
from easytext.trainer.grad_rescaled import GradRescaled
from easytext.trainer import Record
from easytext.trainer.trainer_callback import TrainerCallback
from easytext.distributed import Distributed
from easytext.distributed import DistributedDataParallelParameter, ProcessGroupParameter


class NerModelOutputs:
    """
    Ner Model Outputs
    """

    def __init__(self, logits, mask, crf):
        """
        Ner 模型的输出
        :param logits: logits 输出
        :param mask: mask
        :param crf: 模型中的 crf 输出出来，用来进行 loss 以及 viterbi 解码
        """
        self.logits = logits
        self.mask = mask
        self.crf = crf


class Trainer(TrainerCallback, Distributed):
    """
    训练器
    """
    _TRAIN = 0
    _EVALUATE = 1

    def __init__(self,
                 serialize_dir: str,
                 num_epoch: int,
                 model: Model,
                 loss: Loss,
                 metrics: ModelMetricAdapter,
                 optimizer_factory: OptimizerFactory,
                 device: torch.device,
                 is_distributed: bool,
                 lr_scheduler_factory: LRSchedulerFactory = None,
                 grad_scaled: GradRescaled = None,
                 patient: int = None,
                 num_check_point_keep: int = None,
                 trainer_callback: Union[TrainerCallback, List[TrainerCallback], None] = None,
                 distributed_data_parallel_parameter: DistributedDataParallelParameter = None
                 ):
        """
        训练器初始化
        :param num_epoch: 训练的 epoch 数量
        :param model: 要训练的模型
        :param loss: 模型的 loss function
        :param metrics: 模型的指标计算
        :param optimizer_factory: 模型的优化器的创建工厂。为什么不直接使用优化器？是因为, 优化器的创建依赖于 model, 所以
        这里传递的参数 optimizer factory, 避免使用者在 trainer 外面生成 optimizer, 导致在 trainer 内 optimizer 依赖于
        model 的参数产生问题。典型问题是: 设置 cuda.
        :param device: 训练时所依赖的 device
        :param is_distributed: 当前 trainer 是否是在多 GPU 环境下使用
        :param serialize_dir: 训练存储的文件路径
        :param patient: early stopping 的 patient. 如果是 `None`, 将不会进行 early stopping;
        否则, 当前训练的指标超出了 patient 个 epoch 将会 early stopping.
        :param num_check_point_keep: checkpoint 保留的数量。如果是 `None` 则全部保留;
        否则，保留 num_check_point_keep 个checkpoint.
        :param trainer_callback: 训练中的回调。可以是 List, 如果是 List, 怎按照顺序逐个执行.
            当 Trainer.is_distributed == True, 会去查看 trainer_back.is_distributed, True 表示
        :param distributed_data_parallel_parameter: DistributedDataParallel 用到的参数, 目前只支持设置 find_unused_parameters
        """

        self._device = device
        self._loss = loss
        self._metrics = metrics
        self._optimizer_factory = optimizer_factory

        self._grad_scaled = grad_scaled

        self._serialize_dir = serialize_dir
        self._metric_tracker = MetricTracker(patient=patient)
        self._num_check_point_keep = num_check_point_keep
        self._num_epoch = num_epoch
        self._current_epoch: Optional[int] = None
        self._trainer_callback = trainer_callback

        self._is_distributed = is_distributed

        if self.is_distributed:
            self._distributed_func_wrapper = DistributedFuncWrapper(dst_rank=0)

            self._ddp = distributed_data_parallel_parameter \
                        or DistributedDataParallelParameter(find_unused_parameters=False)
        else:
            self._distributed_func_wrapper = DistributedFuncWrapper(dst_rank=None)

        self._model = model
        if self.is_distributed:

            assert self._device.type != "cpu", f"多 GPU 训练, device 不能是 cpu"

            torch.cuda.set_device(self._device)
            self._model.cuda(self._device)
            self._optimizer = optimizer_factory.create(self._model)

            self._model = DistributedDataParallel(module=self._model,
                                                  device_ids=[self._device],
                                                  output_device=self._device,
                                                  find_unused_parameters=self._ddp.find_unused_parameters)
        else:
            self._model = self._model.to(self._device)
            self._optimizer = self._optimizer_factory.create(self._model)

        if lr_scheduler_factory is not None:
            self._lr_scheduler = lr_scheduler_factory.create(optimizer=self.optimizer,
                                                             model=self.model)
        else:
            self._lr_scheduler = None

        self._check_distributed()

    def _check_distributed(self):
        """
        检查 metric 的合法性, 如果非法会抛出异常
        :return:
        """

        if self._trainer_callback is not None:
            assert self._trainer_callback.is_distributed == self.is_distributed, \
                f"当前 trainer_callback is_distributed: {self._trainer_callback.is_distributed} " \
                f"与 trainer is_distributed:{self.is_distributed} 不相等"

    @property
    def is_distributed(self) -> bool:
        return self._is_distributed

    @property
    def model(self):
        return self._model

    @property
    def loss(self):
        return self._loss

    @property
    def metrics(self):
        return self._metrics

    @property
    def metric_tracker(self):
        return self._metric_tracker

    @property
    def optimizer(self):
        return self._optimizer

    @property
    def serialize_dir(self):
        return self._serialize_dir

    @property
    def num_epoch(self):
        return self._num_epoch

    @property
    def current_epoch(self):
        return self._current_epoch

    def save_checkpoint(self,
                        epoch: int) -> "Trainer":
        """
        保存 checkpoint 到指定的路径下。
        :param epoch: 保存的 epoch
        :return: self
        """

        # 创建 checkpoint dir

        saved_dir = os.path.join(self._serialize_dir, f"checkpoint_epoch_{epoch}")

        if not os.path.isdir(saved_dir):
            os.makedirs(saved_dir)

        # 模型保存
        model_file_path = os.path.join(saved_dir, "model.pt")
        torch.save(self._model.state_dict(), model_file_path)

        # 优化器保存
        optimizer_file_path = os.path.join(saved_dir, "optimizer.pt")
        torch.save(self._optimizer.state_dict(), optimizer_file_path)

        # 保存lr scheduler

        lr_scheduler_file_path = os.path.join(saved_dir, "lr_scheduler.pt")

        if self._lr_scheduler is None:
            torch.save(None, lr_scheduler_file_path)
        else:
            torch.save(self._lr_scheduler.state_dict(), lr_scheduler_file_path)

        # metric 保存
        # 这里保存的是当前 epoch 的 metric, 在最外面还会保存一份完整的 metric tracker
        metric_file_path = os.path.join(saved_dir, "metric.json")

        metric = self._metric_tracker[epoch]
        with open(metric_file_path, mode="w", encoding="utf-8") as f:
            f.write(f"{json2str(metric, indent=2)}\n")

        # metric tracker 保存
        metric_tracker_file_path = os.path.join(self._serialize_dir, "metric_tracker.json")
        self._metric_tracker.save(metric_tracker_file_path)

        # save best
        if self._metric_tracker.best().epoch == epoch:
            # 保存当前为best
            best_dir = os.path.join(self._serialize_dir, "best")

            if os.path.isdir(best_dir):
                # 已经存在，移动到 backup
                best_bak_dir = os.path.join(self._serialize_dir, "best_bak")

                if os.path.isdir(best_bak_dir):  # 删除 bak
                    shutil.rmtree(best_bak_dir)

                # 将 best 的 备份到 best_bak
                shutil.move(best_dir, best_bak_dir)

            shutil.copytree(saved_dir, best_dir)

        # 删除keep last之外的
        if self._num_check_point_keep is not None:
            for removed_epoch in range(1, epoch - self._num_check_point_keep + 1):
                removed_epoch_dir = os.path.join(self._serialize_dir, f"checkpoint_epoch_{removed_epoch}")

                if os.path.isdir(removed_epoch_dir):
                    shutil.rmtree(removed_epoch_dir)

        return self

    @staticmethod
    def _find_last_epoch(serialize_dir: str):
        """
        寻找最后的 epoch
        :param serialize_dir: serialize 目录
        :return:
        """
        # 找到 last epoch
        last_epoch = None
        for file_name in os.listdir(serialize_dir):

            dir_path = os.path.join(serialize_dir, file_name)

            if os.path.isdir(dir_path):
                parts = file_name.split("_")
                if len(parts) == 3 and parts[0] == "checkpoint" and parts[1] == "epoch":
                    epoch = int(parts[2])

                    if last_epoch is None:
                        last_epoch = epoch
                    else:
                        if epoch > last_epoch:
                            last_epoch = epoch
        return last_epoch

    def load_checkpoint(self,
                        serialize_dir: str) -> "Trainer":
        """
        载入 check point
        :param serialize_dir: 保存的路径
        :return: self
        """

        last_epoch = Trainer._find_last_epoch(serialize_dir=serialize_dir)

        if last_epoch is not None:
            self._current_epoch = last_epoch

            if self.is_distributed:

                if self._distributed_func_wrapper is not None \
                        and self._distributed_func_wrapper.dst_rank == TorchDist.get_rank():
                    logging.info(f"Load checkpoint, 当前 epoch: {last_epoch}")
            else:
                logging.info(f"Load checkpoint, 当前 epoch: {last_epoch}")

            saved_dir = os.path.join(serialize_dir, f"checkpoint_epoch_{last_epoch}")

            model_file_path = os.path.join(saved_dir, "model.pt")
            self._model.load_state_dict(torch.load(model_file_path))

            if self.is_distributed:

                if self._distributed_func_wrapper is not None \
                        and self._distributed_func_wrapper.dst_rank == TorchDist.get_rank():
                    logging.info(f"last epoch{last_epoch}, loaded: {self._model.state_dict()}")
            else:
                logging.info(f"last epoch{last_epoch}, loaded: {self._model.state_dict()}")

            optimizer_file_path = os.path.join(saved_dir, "optimizer.pt")
            self._optimizer.load_state_dict(torch.load(optimizer_file_path))

            lr_scheduler_file_path = os.path.join(saved_dir, "lr_scheduler.pt")

            lr_state_dict = torch.load(lr_scheduler_file_path)
            if lr_state_dict is None:
                self._lr_scheduler = None
            else:
                self._lr_scheduler.load_state_dict(lr_state_dict)

            metric_tracker_file_path = os.path.join(serialize_dir, "metric_tracker.json")
            self._metric_tracker = MetricTracker.from_file(metric_tracker_file_path)
        else:
            raise RuntimeError(f"最后保存的epoch数据没有在 {self._serialize_dir} 中找到!")

        return self

    def _train_or_evaluate(self,
                           phrase: int,
                           data_loader: DataLoader) -> float:

        total_loss = 0.

        self._metrics.reset()

        if phrase == Trainer._TRAIN:
            self._model.train()
        elif phrase == Trainer._EVALUATE:
            self._model.eval()
        else:
            raise RuntimeError(f"phrase: {phrase} 应该是 {Trainer._TRAIN} 或 {Trainer._EVALUATE}")

        with torch.set_grad_enabled(phrase == Trainer._TRAIN):

            if self.is_distributed:

                tqdm_disable = True

                if self._distributed_func_wrapper is not None \
                    and self._distributed_func_wrapper.dst_rank == TorchDist.get_rank():
                    tqdm_disable = False
            else:
                tqdm_disable = False

            for model_inputs in tqdm(data_loader, disable=tqdm_disable):
                model_inputs: ModelInputs = model_inputs

                batch_size, batch_inputs, labels \
                    = model_inputs.batch_size, model_inputs.model_inputs, model_inputs.labels

                # 设置到 cuda 训练
                if self._device.type == "cuda":  # 仅仅处理 GPU, 默认使用 CPU
                    batch_inputs = cuda_util.cuda(batch_inputs, cuda_device=self._device)
                    labels = cuda_util.cuda(labels, cuda_device=self._device)

                outputs = self._model(**batch_inputs)

                batch_loss: torch.Tensor = self._loss(outputs, labels)

                if phrase == Trainer._TRAIN:
                    self._optimizer.zero_grad()
                    batch_loss.backward()

                    # 反向传播之后修订梯度
                    if self._grad_scaled is not None:
                        self._grad_scaled(self._model)

                    self._optimizer.step()

                total_loss += batch_loss.detach().item() * batch_size

                batch_metrics, target_metric = self._metrics(model_outputs=outputs, golden_labels=labels)

                if self.is_distributed:

                    if self._distributed_func_wrapper is not None \
                            and self._distributed_func_wrapper.dst_rank == TorchDist.get_rank():
                        logging.info(f"Epoch: {self._current_epoch}, batch loss: {batch_loss}"
                                     f"batch metrics: {json2str(batch_metrics)}, "
                                     f"target metric: {json2str(target_metric)}")
                else:
                    logging.info(f"Epoch: {self._current_epoch}, batch loss: {batch_loss},"
                                 f"batch metrics: {json2str(batch_metrics)}, "
                                 f"target metric: {json2str(target_metric)}")

        # total_loss = total_loss / total_num 这是合理的 loss, 因为所有的 total_num 是一样的所以，没有必要再除以一次了
        return total_loss

    def recovery_train(self,
                       train_data_loader: DataLoader,
                       validation_data_loader: DataLoader):
        """
        恢复训练，是指从上次中断的位置重新开始训练
        :param train_data_loader: 训练数据集
        :param validation_data_loader:
        :return:
        """

        self.load_checkpoint(self.serialize_dir)

        if self.is_distributed:
            TorchDist.barrier()

        self._train(train_data_loader=train_data_loader,
                    validation_data_loader=validation_data_loader)

        if self.is_distributed:
            TorchDist.barrier()

    def evaluate(self,
                 validation_data_loader: DataLoader) -> float:
        """
        评估验证集
        :param validation_data_loader: 验证集data loader
        :return: loss 结果, 以及当前数量
        """
        loss = self._train_or_evaluate(phrase=Trainer._EVALUATE,
                                       data_loader=validation_data_loader)
        if self.is_distributed:
            TorchDist.barrier()

        return loss

    def _is_serialize_empty(self):
        """
        判断 serialize dir 是否是空的，会忽略 隐藏文件
        :return: True: 空文件夹; False: 非空文件夹
        """
        if not os.path.isdir(self._serialize_dir):
            raise RuntimeError(f"保存路径是无效的路径: {self._serialize_dir} ")

        is_empty = True
        for name in os.listdir(self._serialize_dir):
            if not name.startswith("."):
                is_empty = False
                break
        return is_empty

    def _check_data_loader_validity(self, data_loader: DataLoader):
        """
        检查 data loader 是否有效
        :param data_loader: data loader
        :return:
        """

        if self.is_distributed:
            assert isinstance(data_loader.sampler, DistributedSampler), \
                f"data_loader.sampler 必须是 DistributedSampler 实例"

    def train(self,
              train_data_loader: DataLoader,
              validation_data_loader: DataLoader) -> None:
        """
        模型训练
        :param train_data_loader: 训练集 data loader
        :param validation_data_loader: 验证集 data loader
        :return:
        """

        if not self._is_serialize_empty():
            raise RuntimeError(f"新训练，请清空保存文件件: {self._serialize_dir}")

        if self.is_distributed:
            TorchDist.barrier()

        self._check_data_loader_validity(train_data_loader)
        self._check_data_loader_validity(validation_data_loader)

        self._train(train_data_loader=train_data_loader,
                    validation_data_loader=validation_data_loader)

        if self.is_distributed:
            TorchDist.barrier()

    def _train(self,
               train_data_loader: DataLoader,
               validation_data_loader: DataLoader) -> None:
        """
        模型训练
        :param train_data_loader: 训练集 data loader
        :param validation_data_loader: 验证集 data loader
        :return:
        """

        if self._current_epoch is None:
            start_epoch = 1
        else:
            start_epoch = self._current_epoch

        record = Record()

        for epoch in range(start_epoch, self._num_epoch + 1):
            record.epoch = epoch

            self._current_epoch = epoch

            if self.is_distributed:

                if self._distributed_func_wrapper is not None \
                        and self._distributed_func_wrapper.dst_rank == TorchDist.get_rank():
                    logging.info(f"Start train epoch: {self._current_epoch}")
            else:
                logging.info(f"Start train epoch: {self._current_epoch}")

            self.on_train_epoch_start(trainer=self, record=record)

            train_loss = self._train_or_evaluate(phrase=Trainer._TRAIN, data_loader=train_data_loader)

            if self.is_distributed:
                train_loss = Sync.sync(train_loss, device=self._device, op=ReduceOp.SUM)

            record.epoch_train_loss = train_loss

            if self.is_distributed:
                sync_data, op = self._metrics.to_synchronized_data()
                sync_data = Sync.sync(sync_data, device=self._device, op=op)

                self._metrics.from_synchronized_data(sync_data=sync_data, reduce_op=op)

            # 输出metrics
            train_metric_dict, train_target_metric = self._metrics.metric

            record.train_metric = train_metric_dict
            record.train_target_metric = train_target_metric
            
            if self.is_distributed:

                if self._distributed_func_wrapper is not None \
                        and self._distributed_func_wrapper.dst_rank == TorchDist.get_rank():
                    logging.info(f"Train epoch: {epoch}, "
                                 f"loss: {train_loss}, "
                                 f"target metric: {train_target_metric.name}:{train_target_metric.value},"
                                 f"metrics: {json2str(train_metric_dict)}")
            else:
                logging.info(f"Train epoch: {epoch}, "
                             f"loss: {train_loss}, "
                             f"target metric: {train_target_metric.name}:{train_target_metric.value},"
                             f"metrics: {json2str(train_metric_dict)}")

            self.on_train_epoch_stop(trainer=self, record=record)

            self.on_evaluate_validation_epoch_start(trainer=self, record=record)

            validation_loss = self.evaluate(validation_data_loader=validation_data_loader)

            if self.is_distributed:
                validation_loss = Sync.sync(validation_loss, device=self._device, op=ReduceOp.SUM)

            record.epoch_validation_loss = validation_loss

            if self.is_distributed:
                sync_data, op = self._metrics.to_synchronized_data()
                sync_data = Sync.sync(sync_data, device=self._device, op=op)
                self._metrics.from_synchronized_data(sync_data=sync_data, reduce_op=op)

            validation_metric_dict, validation_target_metric = self._metrics.metric

            record.validation_metric = validation_metric_dict
            record.validation_target_metric = validation_target_metric

            self._metric_tracker.add_metric(epoch=epoch,
                                            train_metric=train_metric_dict,
                                            train_model_target_metric=train_target_metric,
                                            validation_metric=validation_metric_dict,
                                            validation_model_target_metric=validation_target_metric)

            if self.is_distributed:

                if self._distributed_func_wrapper is not None \
                        and self._distributed_func_wrapper.dst_rank == TorchDist.get_rank():
                    logging.info(f"Evaluate Valid epoch: {epoch}, loss: {validation_loss}, "
                                 f"target metric: {validation_target_metric.name}:{validation_target_metric.value} "
                                 f"metrics: {json2str(validation_metric_dict)}")
            else:
                logging.info(f"Evaluate Valid epoch: {epoch}, loss: {validation_loss}, "
                             f"target metric: {validation_target_metric.name}:{validation_target_metric.value} "
                             f"metrics: {json2str(validation_metric_dict)}")

            self.on_evaluate_validation_epoch_stop(trainer=self, record=record)

            # 设置 lr scheduler
            # 注意这样设置会有点问题，对于某些 scheduler step 需要参数, 例如: ReduceLROnPlateau
            # 这种情况，暂时不处理, 如果处理需要重新重构，或者简单实用 isinstance ReduceLROnPlateau 来处理
            # 这里简单实用 isinstance 处理。必须指出，ReduceLROnPlateau 的基类是 object, 这也是多少有些问题。

            if self._lr_scheduler is not None:
                if isinstance(self._lr_scheduler, ReduceLROnPlateau):
                    self._lr_scheduler.step(metrics=validation_loss, epoch=epoch)
                else:
                    self._lr_scheduler.step(epoch=epoch)

            self._distributed_func_wrapper(self.save_checkpoint, epoch=epoch)

            if self._metric_tracker.early_stopping(epoch):
                if self.is_distributed:

                    if self._distributed_func_wrapper is not None \
                            and self._distributed_func_wrapper.dst_rank == TorchDist.get_rank():
                        logging.info(f"Epoch: {epoch}, early stopping!")
                else:
                    logging.info(f"Epoch: {epoch}, early stopping!")
                break

        self.on_training_complete(trainer=self, record=record)

    def on_train_epoch_start(self, trainer: "Trainer", record: Record) -> None:

        if self.is_distributed:

            if self._distributed_func_wrapper is not None \
                    and self._distributed_func_wrapper.dst_rank == TorchDist.get_rank():
                logging.info(f"on_train_epoch_start: {record.epoch}")
        else:
            logging.info(f"on_train_epoch_start: {record.epoch}")

        if self._trainer_callback is not None:
            self._trainer_callback.on_train_epoch_start(trainer=trainer,
                                                        record=record)

    def on_train_epoch_stop(self, trainer: "Trainer", record: Record) -> None:

        if self.is_distributed:

            if self._distributed_func_wrapper is not None \
                    and self._distributed_func_wrapper.dst_rank == TorchDist.get_rank():
                logging.info(f"on_train_epoch_stop: {record.epoch}")
        else:
            logging.info(f"on_train_epoch_stop: {record.epoch}")

        if self._trainer_callback is not None:
            self._trainer_callback.on_train_epoch_stop(trainer=trainer,
                                                       record=record)

    def on_evaluate_validation_epoch_start(self, trainer: "Trainer", record: Record) -> None:

        if self.is_distributed:

            if self._distributed_func_wrapper is not None \
                    and self._distributed_func_wrapper.dst_rank == TorchDist.get_rank():
                logging.info(f"on_evaluate_epoch_start: {record.epoch}")
        else:
            logging.info(f"on_evaluate_epoch_start: {record.epoch}")

        if self._trainer_callback is not None:
            self._trainer_callback.on_evaluate_validation_epoch_start(trainer=trainer,
                                                                      record=record)

    def on_evaluate_validation_epoch_stop(self, trainer: "Trainer", record: Record) -> None:

        if self.is_distributed:

            if self._distributed_func_wrapper is not None \
                    and self._distributed_func_wrapper.dst_rank == TorchDist.get_rank():
                logging.info(f"on_evaluate_epoch_stop: {record.epoch}")
        else:
            logging.info(f"on_evaluate_epoch_stop: {record.epoch}")

        if self._trainer_callback is not None:
            self._trainer_callback.on_evaluate_validation_epoch_stop(trainer=trainer,
                                                                     record=record)

    def on_training_complete(self, trainer: "Trainer", record: Record) -> None:

        if self.is_distributed:

            if self._distributed_func_wrapper is not None \
                    and self._distributed_func_wrapper.dst_rank == TorchDist.get_rank():
                logging.info(f"on_training_complete: {record.epoch}")
        else:
            logging.info(f"on_training_complete: {record.epoch}")

        if self._trainer_callback is not None:
            self._trainer_callback.on_training_complete(trainer=trainer,
                                                        record=record)



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
"""
模型

Authors: panxu(panxu@baidu.com)
Date:    2020/05/18 10:08:00
"""

from typing import Any, Dict
import torch
from torch.nn import Module

from .model_outputs import ModelOutputs


class Model(Module):
    """
    模型
    """

    def __init__(self):
        super().__init__()

    def reset_parameters(self):
        raise NotImplementedError()

    def forward(self, *input: Any, **kwargs: Any) -> ModelOutputs:
        pass





#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
时间消耗 callback

Authors: PanXu
Date:    2020/10/16 15:48:00
"""
import logging
import time
from typing import Optional

from torch.utils.tensorboard import SummaryWriter

from easytext.trainer import Record
from easytext.trainer.trainer_callback import TrainerCallback
from easytext.trainer.tensorboard_manager import MainTagManager
from easytext.utils.distributed.distributed_util import DistributedFuncWrapper


class TimeConsumeCallback(TrainerCallback):
    """
    时间消耗 callback 用来统计不同阶段的时间, 时间单位是 毫秒 ms
    """

    def __init__(self, tensorboard_summary_writer: SummaryWriter,
                 is_distributed: bool = False,
                 distributed_func_wrapper: Optional[DistributedFuncWrapper] = None):
        self._writer = tensorboard_summary_writer
        self._train_epoch_start_time: int = None
        self._evaluate_epoch_start_time: int = None
        self._is_distributed = is_distributed

        if self._is_distributed:
            self._distributed_func_wrapper = distributed_func_wrapper or DistributedFuncWrapper(dst_rank=0)
        else:
            self._distributed_func_wrapper = DistributedFuncWrapper(dst_rank=None)

    @property
    def is_distributed(self) -> bool:
        return self._is_distributed

    def on_train_epoch_start(self, trainer: "Trainer", record: Record) -> None:

        self._distributed_func_wrapper(
            self._on_train_epoch_start, trainer=trainer, record=record
        )

    def _on_train_epoch_start(self, trainer: "Trainer", record: Record) -> None:
        self._train_epoch_start_time = time.time()

    def on_train_epoch_stop(self, trainer: "Trainer", record: Record) -> None:

        self._distributed_func_wrapper(
            self._on_train_epoch_stop, trainer=trainer, record=record
        )

    def _on_train_epoch_stop(self, trainer: "Trainer", record: Record) -> None:
        stop_time = time.time()
        used_time = stop_time - self._train_epoch_start_time
        used_time = int(1000 * used_time)
        logging.info(f"Train epoch {record.epoch} used: {used_time}ms")
        self._writer.add_scalars(main_tag=MainTagManager.EPOCH_USED_TIME,
                                 tag_scalar_dict={"train": used_time},
                                 global_step=record.epoch)

    def on_evaluate_validation_epoch_start(self, trainer: "Trainer", record: Record) -> None:

        self._distributed_func_wrapper(
            self._on_evaluate_validation_epoch_start, trainer=trainer, record=record)

    def _on_evaluate_validation_epoch_start(self, trainer: "Trainer", record: Record) -> None:
        self._evaluate_epoch_start_time = time.time()

    def on_evaluate_validation_epoch_stop(self, trainer: "Trainer", record: Record) -> None:

        self._distributed_func_wrapper(
            self._on_evaluate_validation_epoch_stop, trainer=trainer, record=record
        )

    def _on_evaluate_validation_epoch_stop(self, trainer: "Trainer", record: Record) -> None:
        stop_time = time.time()
        used_time = stop_time - self._evaluate_epoch_start_time
        used_time = int(1000 * used_time)
        logging.info(f"Evaluate validation epoch {record.epoch} used: {used_time}ms")
        self._writer.add_scalars(main_tag=MainTagManager.EPOCH_USED_TIME,
                                 tag_scalar_dict={"validation": used_time},
                                 global_step=record.epoch)

    def on_training_complete(self, trainer: "Trainer", record: Record) -> None:
        # 注意这里要 flush 不然可能会导致没有输出

        return self._distributed_func_wrapper(
            self._on_training_complete, trainer=trainer, record=record
        )

    def _on_training_complete(self, trainer: "Trainer", record: Record) -> None:
        # 注意这里要 flush 不然可能会导致没有输出
        self._writer.flush()



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 Baidu.com, Inc. All Rights Reserved
#
"""
测试 metric tracker

Authors: panxu(panxu@baidu.com)
Date:    2020/05/28 14:46:00
"""
import os
from easytext.tests import ASSERT
from easytext.tests import ROOT_PATH

from easytext.trainer import MetricTracker
from easytext.metrics import ModelTargetMetric

METRICS = [{"epoch": 1,
            "train_metric": {"acc": 0.81},
            "train_model_target_metric": ModelTargetMetric(metric_name="acc", metric_value=0.81),
            "validation_metric": {"acc": 0.46},
            "validation_model_target_metric": ModelTargetMetric(metric_name="acc", metric_value=0.46)},
           {"epoch": 2,
            "train_metric": {"acc": 0.83},
            "train_model_target_metric": ModelTargetMetric(metric_name="acc", metric_value=0.83),
            "validation_metric": {"acc": 0.48},
            "validation_model_target_metric": ModelTargetMetric(metric_name="acc", metric_value=0.48)},
           {"epoch": 3,
            "train_metric": {"acc": 0.85},
            "train_model_target_metric": ModelTargetMetric(metric_name="acc", metric_value=0.85),
            "validation_metric": {"acc": 0.60},
            "validation_model_target_metric": ModelTargetMetric(metric_name="acc", metric_value=0.60)},
           {"epoch": 4,
            "train_metric": {"acc": 0.89},
            "train_model_target_metric": ModelTargetMetric(metric_name="acc", metric_value=0.89),
            "validation_metric": {"acc": 0.44},
            "validation_model_target_metric": ModelTargetMetric(metric_name="acc", metric_value=0.44)},
           {"epoch": 5,
            "train_metric": {"acc": 0.92},
            "train_model_target_metric": ModelTargetMetric(metric_name="acc", metric_value=0.92),
            "validation_metric": {"acc": 0.39},
            "validation_model_target_metric": ModelTargetMetric(metric_name="acc", metric_value=0.39)}
           ]


def test_metric_tracker_best():
    """
    测试 metric tracker
    :return:
    """
    metric_tracker = MetricTracker(patient=None)

    for metric in METRICS:
        metric_tracker.add_metric(**metric)

    expect = {"epoch": 3,
              "train_metric": {"acc": 0.85},
              "train_model_target_metric": ModelTargetMetric(metric_name="acc", metric_value=0.85),
              "validation_metric": {"acc": 0.60},
              "validation_model_target_metric": ModelTargetMetric(metric_name="acc", metric_value=0.60)}

    best = metric_tracker.best()
    ASSERT.assertEqual(expect["epoch"], best.epoch)

    ASSERT.assertDictEqual(expect["train_metric"], best.train_metric)
    ASSERT.assertDictEqual(expect["validation_metric"], best.validation_metric)
    ASSERT.assertEqual(expect["train_model_target_metric"].name, best.train_model_target_metric.name)
    ASSERT.assertEqual(expect["train_model_target_metric"].value, best.train_model_target_metric.value)
    ASSERT.assertEqual(expect["validation_model_target_metric"].name, best.validation_model_target_metric.name)
    ASSERT.assertEqual(expect["validation_model_target_metric"].value, best.validation_model_target_metric.value)


def test_metric_tracker_patient():
    metric_tracker = MetricTracker(patient=1)

    for metric in METRICS:
        metric_tracker.add_metric(**metric)

        if metric["epoch"] > 4:
            ASSERT.assertTrue(metric_tracker.early_stopping(metric["epoch"]))
        else:
            ASSERT.assertFalse(metric_tracker.early_stopping(metric["epoch"]))

        if metric_tracker.early_stopping(metric["epoch"]):
            break

    expect = {"epoch": 3,
              "train_metric": {"acc": 0.85},
              "train_model_target_metric": ModelTargetMetric(metric_name="acc", metric_value=0.85),
              "validation_metric": {"acc": 0.60},
              "validation_model_target_metric": ModelTargetMetric(metric_name="acc", metric_value=0.60)}

    best = metric_tracker.best()
    ASSERT.assertEqual(expect["epoch"], best.epoch)

    ASSERT.assertDictEqual(expect["train_metric"], best.train_metric)
    ASSERT.assertDictEqual(expect["validation_metric"], best.validation_metric)
    ASSERT.assertEqual(expect["train_model_target_metric"].name, best.train_model_target_metric.name)
    ASSERT.assertEqual(expect["train_model_target_metric"].value, best.train_model_target_metric.value)
    ASSERT.assertEqual(expect["validation_model_target_metric"].name, best.validation_model_target_metric.name)
    ASSERT.assertEqual(expect["validation_model_target_metric"].value, best.validation_model_target_metric.value)


def test_metric_tracker_save_and_load():
    metric_tracker = MetricTracker(patient=1)

    for metric in METRICS:
        metric_tracker.add_metric(**metric)

        if metric["epoch"] > 4:
            ASSERT.assertTrue(metric_tracker.early_stopping(metric["epoch"]))
        else:
            ASSERT.assertFalse(metric_tracker.early_stopping(metric["epoch"]))

        if metric_tracker.early_stopping(metric["epoch"]):
            break

    saved_file_path = os.path.join(ROOT_PATH, "data/easytext/tests/trainer/metric_tracker.json")

    metric_tracker.save(saved_file_path)

    loaded_metric_tracker = MetricTracker.from_file(saved_file_path)

    best = metric_tracker.best()
    loaded_best = loaded_metric_tracker.best()
    ASSERT.assertEqual(best.epoch, loaded_best.epoch)

    ASSERT.assertDictEqual(best.train_metric, loaded_best.train_metric)
    ASSERT.assertDictEqual(best.validation_metric, loaded_best.validation_metric)
    ASSERT.assertEqual(best.train_model_target_metric.name, loaded_best.train_model_target_metric.name)
    ASSERT.assertEqual(best.train_model_target_metric.value, loaded_best.train_model_target_metric.value)
    ASSERT.assertEqual(best.validation_model_target_metric.name, loaded_best.validation_model_target_metric.name)
    ASSERT.assertEqual(best.validation_model_target_metric.value, loaded_best.validation_model_target_metric.value)


#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
注册器

Authors: PanXu
Date:    2020/10/27 15:11:00
"""

from typing import Dict, Type, TypeVar

T = TypeVar('T')


class Registry:
    """
    全局注册器，单件模式
    """

    __instance = None

    def __new__(cls, *args, **kwargs):
        if not cls.__instance:
            cls.__instance = super().__new__(cls)

            # 类注册表
            cls._class_registry: Dict[str, Dict[str, Type[T]]] = dict()

            # 对象注册表
            cls._object_registry: Dict[str, Dict[str, T]] = dict()

        return cls.__instance

    def __init__(self):
        # 这里不应该进行任何初始化行为
        pass

    def register_class(self, cls: Type[T], name_space: str, name: str, is_allowed_exist: bool = False) -> "Registry":
        """
        注册类
        :param cls: 要注册的 类
        :param name_space: 命名空间
        :param name: 配置文件中的名字
        :param is_allowed_exist: True: 如果名字和命名空间一样, 会覆盖存在的类; False: 会报错
        :return:
        """

        name_space_dict = self._class_registry.get(name_space, dict())

        if name in name_space_dict and not is_allowed_exist:
            raise RuntimeError(f"{name_space}.{name} {cls} 已经存在!")

        name_space_dict[name] = cls

        self._class_registry[name_space] = name_space_dict
        return self

    def find_class(self, name_space: str, name: str) -> Type[T]:
        """
        根据 name space 或 name 找到 class 或 function
        :param name_space: 类在config文件中的命名空间
        :param name: 泪在 config 文件中的 类型
        :return:
        """
        return self._class_registry.get(name_space, dict()).get(name, None)

    def register_object(self, obj: T, name: str) -> "Registry":
        """
        注册 object
        :param obj: 实例对象
        :param name: 实例对象的名字
        :return:
        """

        if name in self._object_registry:
            raise RuntimeError(f"{name} object 已经存在!")

        self._object_registry[name] = obj

        return self

    def find_object(self, name: str) -> T:
        """
        根据配置文件中的名字找到对象
        :param name: 对象的名字
        :return:
        """
        return self._object_registry[name]

    def clear_objects(self) -> "Registry":
        """
        清空注册表
        :return:
        """
        self._object_registry.clear()

        return self





#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
trainer 回调

Authors: PanXu
Date:    2020/10/13 14:36:00
"""

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from easytext.trainer import Trainer

from easytext.trainer.record import Record
from easytext.distributed import Distributed


class TrainerCallback(Distributed):
    """
    trainer callbeck 接口
    """

    @property
    def is_distributed(self) -> bool:
        raise NotImplementedError()

    def on_train_epoch_start(self, trainer: "Trainer", record: Record) -> None:
        """
        在每一个 epoch 训练开始时候会调用
        :param trainer: 训练器
        :param record: 训练中的记录数据
        :return:
        """
        raise NotImplemented()

    def on_train_epoch_stop(self, trainer: "Trainer", record: Record) -> None:
        """
        在每一个 epoch 训练结束时候会调用
        :param trainer: 训练器
        :param record: 训练中的记录数据
        :return:
        """
        raise NotImplemented()

    def on_evaluate_validation_epoch_start(self, trainer: "Trainer", record: Record) -> None:
        """
        在每一个 epoch evaluate 验证集开始的时候会调用
        :param trainer: 训练器
        :param record: 训练中的记录数据
        :return:
        """
        raise NotImplemented()

    def on_evaluate_validation_epoch_stop(self, trainer: "Trainer", record: Record) -> None:
        """
        在每一个 epoch evaluate 验证集结束的时候会调用
        :param trainer: 训练器
        :param record: 训练中的记录数据
        :return:
        """
        raise NotImplemented()

    def on_training_complete(self, trainer: "Trainer", record: Record) -> None:
        """
        训练过程全部结束
        :param trainer: 训练器
        :param record: 训练中的记录数据
        :return:
        """
        raise NotImplemented




#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
动态 rnn 处理变长序列，batch first

Authors: PanXu
Date:    2020/10/20 15:28:00
"""

from typing import Dict

import torch
from torch import Tensor, BoolTensor
from torch.nn import Module, LSTM, GRU, RNN, RNNBase
from torch.nn.utils.rnn import pack_padded_sequence, pad_packed_sequence


class DynamicRnnOutput:
    """
    成员变量:
    * last_layer_h_n: 最后一层的最后一个 time step 隐层输出，
                      batch first shape: (batch, hidden_size * num_directions)
    * last_layer_c_n: 最后一层的最后一个 time step cell输出
                      batch first shape: (batch, hidden_size * num_directions)
    * h_n: 所有层的最后一个 time step 的隐层输出，具体参考 pytorch lstm/gru/rnn 文档描述
                      batch first shape: (batch, num_layers * num_directions, hidden_size)
    * c_n: 所有层的最后一个 time step 的cell 输出, 具体参考 pytorch lstm/gru/rnn 文档描述
           对于 rnn 来说，为 None. batch first shape: (batch, num_layers * num_directions, hidden_size)
    * output: 经过 rnn 后的序列输出, 具体参考 pytorch lstm/gru/rnn 文档描述
              batch first shape: (batch, seq_len, num_directions * hidden_size),
    """

    def __init__(self,
                 last_layer_h_n: Tensor,
                 last_layer_c_n: Tensor,
                 h_n: Tensor,
                 c_n: Tensor,
                 sequence_encoding: Tensor):
        """
        :param last_layer_h_n: 最后一层的最后一个 time step 隐层输出，
                batch first shape: (batch, hidden_size * num_directions)
        :param last_layer_c_n: 最后一层的最后一个 time step cell输出
                batch first shape: (batch, hidden_size * num_directions)
        :param h_n: 所有层的最后一个 time step 的隐层输出，具体参考 pytorch lstm/gru/rnn 文档描述
                batch first shape: (batch, num_layers * num_directions, hidden_size)
        :param c_n: 所有层的最后一个 time step 的cell 输出, 具体参考 pytorch lstm/gru/rnn 文档描述
                对于 rnn 来说，为 None. batch first shape: (batch, num_layers * num_directions, hidden_size)
        :param sequence_encoding: 经过 rnn 后的序列输出, 具体参考 pytorch lstm/gru/rnn 文档描述
                batch first shape: (batch, seq_len, num_directions * hidden_size),
        """
        self.last_layer_h_n = last_layer_h_n
        self.last_layer_c_n = last_layer_c_n
        self.h_n = h_n
        self.c_n = c_n
        self.output = sequence_encoding


class DynamicRnn(Module):
    """
    动态 rnn, 注意所有的都是 batch first。
    """
    LSTM = "lstm"
    GRU = "gru"
    RNN = "rnn"

    def __init__(self, rnn: RNNBase):
        super().__init__()

        self.rnn = rnn

        if isinstance(self.rnn, LSTM):
            self.rnn_type = DynamicRnn.LSTM
        elif isinstance(self.rnn, GRU):
            self.rnn_type = DynamicRnn.GRU
        elif isinstance(self.rnn, RNN):
            self.rnn_type = DynamicRnn.RNN
        else:
            raise RuntimeError(f"rnn 类型: {type(self.rnn)} 不属于 "
                               f"{DynamicRnn.LSTM}, {DynamicRnn.GRU}, {DynamicRnn.RNN}")

        assert self.rnn.batch_first, f"rnn batch_first 必须设置为 True"

        self.num_layers = self.rnn.num_layers

    def forward(self, sequence: Tensor, mask: BoolTensor) -> DynamicRnnOutput:
        """
        rnn 执行。特别注意: 所有的都是 batch first
        :param sequence: sequence 序列, shape: (B, seq_len, input_size)
        :param mask: 对 sequence 的 mask, shape: (B, seq_len)
        :return: 解码后的结果，具体参考 DynamicOutput 说明
        """
        assert sequence.dim() == 3, \
            f"sequence shape: {sequence.dim()} 与 (B, seq_len, input_size) 不匹配"

        assert sequence.size(-1) == self.rnn.input_size, \
            f"sequence.size(-1): {sequence.size(-1)} 与 rnn input_size: {self.rnn.input_size} 不相等"

        batch_size = sequence.size(0)
        sequence_length = sequence.size(1)

        sequence_lengths = mask.sum(dim=-1)

        pack = pack_padded_sequence(sequence,
                                    lengths=sequence_lengths,
                                    batch_first=True,
                                    enforce_sorted=False)

        packed_sequence_encoding, last_state = self.rnn(pack)

        encoding, pad_sequence_length = pad_packed_sequence(packed_sequence_encoding,
                                                            batch_first=True,
                                                            padding_value=0.0,
                                                            total_length=sequence_length)

        if self.rnn_type == DynamicRnn.LSTM or self.rnn_type == DynamicRnn.GRU:
            h_n, c_n = last_state
        else:
            h_n = last_state
            c_n = None

        # h_n shape: (num_layers * num_directions, batch, hidden_size)
        # 因为是按照 batch first 来处理的，所以需要进行转换
        # 转换之后的 h_n shape: (batch, num_layers, hidden_size * num_directions), c_n 同样的处理
        h_n = torch.transpose(h_n, 0, 1).contiguous().view(batch_size, self.num_layers, -1)
        last_layer_h_n = h_n[:, -1, :].contiguous().view(batch_size, -1)

        last_layer_c_n = None
        if c_n is not None:
            c_n = torch.transpose(c_n, 0, 1).contiguous().view(batch_size, self.num_layers, -1)
            last_layer_c_n = c_n[:, -1, :].contiguous().view(batch_size, -1)

        return DynamicRnnOutput(last_layer_h_n=last_layer_h_n,
                                last_layer_c_n=last_layer_c_n,
                                h_n=h_n,
                                c_n=c_n,
                                sequence_encoding=encoding)



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
模型输入

Authors: panxu(panxu@baidu.com)
Date:    2020/06/10 09:55:00
"""
from typing import Dict

import torch


class ModelInputs:
    """
    批量的模型输入
    """

    def __init__(self,
                 batch_size: int,
                 model_inputs: Dict[str, torch.Tensor],
                 labels: torch.Tensor = None):
        """
        初始化
        :param batch_size: batch_size
        :param model_inputs: 模型输入，这个参数会被以 **model_inputs 传给模型，所以该字典的key 要与模型参数保持一致,
        相当于 X.
        :param labels: labels 如果是测试集，那么 labels=None. 相当于 Y.
        """
        self.batch_size = batch_size
        self.model_inputs = model_inputs
        self.labels = labels



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2021 PanXu, Inc. All Rights Reserved
#
"""
测试 bio schema 文件

Authors: PanXu
Date:    2021/02/03 10:02:00
"""

from easytext.utils import bio_schema

from easytext.tests import ASSERT


def test_ibo1_to_bio():
    """
    测试 ibo1 转换到 bio
    :return:
    """
    ibo1 = ["I-L1", "I-L1", "O",
            "I-L1", "I-L2", "O",
            "I-L1", "I-L1", "I-L1", "B-L1", "I-L1", "O",
            "B-L1", "I-L1", "O"]

    expect_bio = ["B-L1", "I-L1", "O",
                  "B-L1", "B-L2", "O",
                  "B-L1", "I-L1", "I-L1", "B-L1", "I-L1", "O",
                  "B-L1", "I-L1", "O"]

    bio_sequence = bio_schema.ibo1_to_bio(ibo1)

    ASSERT.assertListEqual(expect_bio, bio_sequence)


def test_bmes_to_bio():
    """
    测试 BMES schema 转换成 bio
    :return:
    """
    bmes = ["B-T", "M-T", "E-T", "O", "S-T", "B-T", "E-T"]
    expect_bio = ["B-T", "I-T", "I-T", "O", "B-T", "B-T", "I-T"]

    bio_sequence_label = bio_schema.bmes_to_bio(bmes)

    ASSERT.assertListEqual(expect_bio, bio_sequence_label)



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2021 PanXu, Inc. All Rights Reserved
#
"""
测试 lattice lstm

Authors: PanXu
Date:    2021/01/21 09:43:00
"""

import torch

from easytext.modules.lattice_lstm import WordLSTMCell, MultiInputLSTMCell

from easytext.tests import ASSERT


def test_word_lstm_cell_with_bias():
    """
    测试 WordLSTMCell
    :return:
    """

    input_size = 2
    hidden_size = 3
    word_lstm_cell = WordLSTMCell(input_size=input_size, hidden_size=hidden_size, bias=True)

    value = list()

    for i in range(input_size):
        value.append([j * 0.37 for j in range(i * hidden_size * 3, (i + 1) * hidden_size * 3)])

    with torch.no_grad():
        word_lstm_cell.weight_ih.copy_(torch.tensor(value, dtype=torch.float))
        torch.nn.init.constant(word_lstm_cell.bias, val=1.0)

    word_input = torch.tensor([[0.2, 0.4]], dtype=torch.float)
    h = torch.tensor([[0.2, 0.11, 0.15]], dtype=torch.float)
    c = torch.tensor([[0.5, 0.6, 0.7]], dtype=torch.float)

    output_c = word_lstm_cell(input_=word_input,
                              hx=(h, c))

    expect_size = (1, hidden_size)
    ASSERT.assertEqual(expect_size, output_c.size())

    expect_output_c = [1.4231, 1.5257, 1.6372]

    for e_i, i in zip(expect_output_c, output_c[0].tolist()):
        ASSERT.assertAlmostEqual(e_i, i, places=3)


def test_word_lstm_cell_without_bias():
    """
    测试 WordLSTMCell
    :return:
    """

    input_size = 2
    hidden_size = 3
    word_lstm_cell = WordLSTMCell(input_size=input_size, hidden_size=hidden_size, bias=False)

    value = list()

    for i in range(input_size):
        value.append([j * 0.37 for j in range(i * hidden_size * 3, (i + 1) * hidden_size * 3)])

    with torch.no_grad():
        word_lstm_cell.weight_ih.copy_(torch.tensor(value, dtype=torch.float))

    word_input = torch.tensor([[0.2, 0.4]], dtype=torch.float)
    h = torch.tensor([[0.2, 0.11, 0.15]], dtype=torch.float)
    c = torch.tensor([[0.5, 0.6, 0.7]], dtype=torch.float)

    output_c = word_lstm_cell(input_=word_input,
                              hx=(h, c))

    expect_size = (1, hidden_size)
    ASSERT.assertEqual(expect_size, output_c.size())

    expect_output_c = [1.3054, 1.4113, 1.5386]

    for e_i, i in zip(expect_output_c, output_c[0].tolist()):
        ASSERT.assertAlmostEqual(e_i, i, places=3)


def test_multi_input_lstm_cell():
    """
    测试 MultiInputLSTMCell
    """

    input_size = 2
    hidden_size = 3

    cell = MultiInputLSTMCell(input_size=input_size, hidden_size=hidden_size, bias=True)

    with torch.no_grad():
        weight_ih_value = list()

        for i in range(input_size):
            weight_ih_value.append([j * 0.37 for j in range(i * hidden_size * 3, (i + 1) * hidden_size * 3)])

        cell.weight_ih.copy_(torch.tensor(weight_ih_value, dtype=torch.float))

        alpha_weight_ih_value = list()

        for i in range(input_size):
            alpha_weight_ih_value.append([j * 0.23 for j in range(i * hidden_size, (i + 1) * hidden_size)])

        cell.alpha_weight_ih.copy_(torch.tensor(alpha_weight_ih_value, dtype=torch.float))

        torch.nn.init.constant(cell.bias, val=1.0)
        torch.nn.init.constant(cell.alpha_bias, val=0.5)

    char_input = torch.tensor([[0.2, 0.4]], dtype=torch.float)

    h = torch.tensor([[0.2, 0.11, 0.15]], dtype=torch.float)
    c = torch.tensor([[0.5, 0.6, 0.7]], dtype=torch.float)

    word_c_input = [torch.tensor([[0.7, 0.5, 0.2]], dtype=torch.float),
                    torch.tensor([[0.3, 0.4, 1.5]], dtype=torch.float)]

    output_hc = cell(input_=char_input,
                     c_input=word_c_input,
                     hx=(h, c))

    expect_size = (1, hidden_size)

    ASSERT.assertEqual(expect_size, output_hc[0].size())
    ASSERT.assertEqual(expect_size, output_hc[1].size())

    expects = [[0.5728, 0.5523, 0.7130], [0.6873, 0.6506, 0.9345]]

    for expect, hc in zip(expects, output_hc):

        for e_i, hc_i in zip(expect, hc[0].tolist()):
            ASSERT.assertAlmostEqual(e_i, hc_i, places=4)


def test_multi_input_lstm_cell_without_bias():
    """
    测试 MultiInputLSTMCell
    """

    input_size = 2
    hidden_size = 3

    cell = MultiInputLSTMCell(input_size=input_size, hidden_size=hidden_size, bias=False)

    with torch.no_grad():
        weight_ih_value = list()

        for i in range(input_size):
            weight_ih_value.append([j * 0.37 for j in range(i * hidden_size * 3, (i + 1) * hidden_size * 3)])

        cell.weight_ih.copy_(torch.tensor(weight_ih_value, dtype=torch.float))

        alpha_weight_ih_value = list()

        for i in range(input_size):
            alpha_weight_ih_value.append([j * 0.23 for j in range(i * hidden_size, (i + 1) * hidden_size)])

        cell.alpha_weight_ih.copy_(torch.tensor(alpha_weight_ih_value, dtype=torch.float))

    char_input = torch.tensor([[0.2, 0.4]], dtype=torch.float)

    h = torch.tensor([[0.2, 0.11, 0.15]], dtype=torch.float)
    c = torch.tensor([[0.5, 0.6, 0.7]], dtype=torch.float)

    word_c_input = [torch.tensor([[0.7, 0.5, 0.2]], dtype=torch.float),
                    torch.tensor([[0.3, 0.4, 1.5]], dtype=torch.float)]

    output_hc = cell(input_=char_input,
                     c_input=word_c_input,
                     hx=(h, c))

    expect_size = (1, hidden_size)

    ASSERT.assertEqual(expect_size, output_hc[0].size())
    ASSERT.assertEqual(expect_size, output_hc[1].size())

    expects = [[0.5356, 0.5204, 0.6862], [0.6855, 0.6490, 0.9451]]

    for expect, hc in zip(expects, output_hc):

        for e_i, hc_i in zip(expect, hc[0].tolist()):
            ASSERT.assertAlmostEqual(e_i, hc_i, places=4)



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
模块相关

Authors: panxu(panxu@baidu.com)
Date:    2020/06/23 11:58:00
"""

from .conditional_random_field import ConditionalRandomField
from .dynamic_rnn import DynamicRnn, DynamicRnnOutput
from .lattice_lstm import LatticeLSTM



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
对参数的相关梯度进行调整

Authors: PanXu
Date:    2020/08/21 09:34:00
"""

from torch.nn import Module
from torch.nn.utils import clip_grad_value_
from easytext.model import Model

from .grad_rescaled import GradRescaled


class GradClipValue(GradRescaled):
    """
    对梯度进行 clip
    """

    def __init__(self, clip_value: float):
        """
        初始化
        :param clip_value: 梯度进行 clip 的 value, 区间是 [-clip_value, clip_value]. clip_value 值大于 0.
        一般会设置为 5 或者 10.
        """
        self._clip_value = clip_value

        assert self._clip_value > 0., f"{self._clip_value} 非法，应该大于 0"

    def __call__(self, model: Module):
        clip_grad_value_(model.parameters(), self._clip_value)




#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
训练启动器

Authors: PanXu
Date:    2020/12/10 10:20:00
"""

from typing import List, Union, Optional, Dict
import logging

import torch
from torch import distributed as TorchDist

from easytext.trainer import Config
from easytext.distributed import ProcessGroupParameter


class Launcher:
    """
    训练启动器
    """

    def __init__(self):
        """
        启动器初始化
        """
        self._devices: Union[List[int], List[str]] = self._init_devices()

    def _init_devices(self) -> Union[List[str], List[int]]:
        """
        返回训练需要用到的设备
        :return: device, 如果是单一设备，返回的也是 List.
            如果是 List[str], ["cuda:0", "cuda:1"]; 如果是 List[int], [0, 1]
        """
        raise NotImplementedError()

    def _init_process_group_parameter(self,
                                      rank: Optional[int]) -> Optional[ProcessGroupParameter]:
        """
        初始化 多GPU/分布式参数
        :param rank: 多 GPU 训练时, 当前进程 id; 单 GPU 或 CPU 该参数为 None.
        :return: 进程组参数
        """
        raise NotImplementedError

    def _start_process(self, rank: Optional[int], world_size: int, devices: Union[List[str], List[int]]) -> None:
        """
        启动训练过程的函数,子类中实现完整的训练过程。一般是指，使用 trainer 进行训练的过程。
        :param rank: 多 GPU 训练时, 当前进程 id; 单 GPU 或 CPU 该参数为 None.
        :param world_size: 全部的训练进程数量, 如果是单 GPU 或 CPU 训练, 那么该值为 1
        :param devices: 训练用的 devices
        :return: None
        """

        process_group_parameter = self._init_process_group_parameter(rank=rank)

        TorchDist.init_process_group(backend=process_group_parameter.backend,
                                     world_size=world_size,
                                     rank=rank,
                                     init_method=process_group_parameter.url)

        device = torch.device(devices[rank])
        self._start(rank=rank, world_size=world_size, device=device)

    def _start(self, rank: Optional[int], world_size: int, device: torch.device) -> None:
        """
        启动训练过程的函数,子类中实现完整的训练过程。一般是指，使用 trainer 进行训练的过程。
        :param rank: 多 GPU 训练时, 当前进程 id; 单 GPU 或 CPU 该参数为 None.
        :param world_size: 全部的训练进程数量, 如果是单 GPU 或 CPU 训练, 那么该值为 1
        :param device: 当前使用的 device
        :return: None
        """
        raise NotImplementedError()

    def _preprocess(self):
        """
        预处理，当多 GPU 训练时候，在主进程中进行的，所以，在训练前一些路径处理等，可以在此处理
        :return:
        """
        pass

    def __call__(self):

        self._preprocess()

        devices = self._devices

        if len(devices) > 1:
            devices_str = ", ".join([str(device) for device in devices])
            logging.info(f"开始在 {devices_str} 上训练...")
            torch.multiprocessing.spawn(fn=self._start_process,
                                        args=(len(devices), devices),
                                        nprocs=len(devices))
        elif len(devices) == 1:
            logging.info(f"开始在 {devices[0]} device 上训练...")
            self._start(rank=None, world_size=1, device=torch.device(devices[0]))
        else:
            logging.info(f"开始在 cpu 上训练...")
            self._start(rank=None, world_size=1, device=torch.device(type="cpu"))




#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
crf label index decoder

Authors: PanXu
Date:    2020/07/05 11:13:00
"""
from typing import List
import torch

from easytext.modules import ConditionalRandomField
from easytext.data import LabelVocabulary
from .label_index_decoder import LabelIndexDecoder


class CRFLabelIndexDecoder(LabelIndexDecoder):
    """
    CRF 的label index decoder, viterbi 算法
    """

    def __init__(self,
                 label_vocabulary: LabelVocabulary,
                 crf: ConditionalRandomField):
        self._crf = crf
        self._label_vocabulary = label_vocabulary

    def __call__(self,
                 logits: torch.Tensor,
                 mask: torch.ByteTensor) -> torch.LongTensor:

        assert logits.dim() == 3, \
            f"logits shape 不匹配, 应该是: (batch_size, seq_len, num_label), 现在是: {logits.dim()}"
        assert logits.size(-1) == self._label_vocabulary.label_size, \
            f"logits.size(-1) 与 label_size 不匹配, 现在是: {logits.size(-1)}"

        assert mask.dim() == 2, f"mask shape 不配, 应该是: (batch_size, seq_len), 现在是: {mask.dim()}"

        sequence_length = logits.size(1)

        best = self._crf.viterbi_tags(logits=logits, mask=mask)

        best_paths, scores = zip(*best)

        label_indices = list()

        for best_path in best_paths:
            best_path: List = best_path
            padding_indices = [self._label_vocabulary.padding_index] * (sequence_length - len(best_path))
            best_path.extend(padding_indices)

            label_indices.append(best_path)

        return torch.tensor(label_indices, dtype=torch.long)



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
对 trainer config 进行测试

Authors: PanXu
Date:    2020/10/29 10:18:00
"""
import os

from easytext.component.register import ComponentRegister
from easytext.trainer import Config

from easytext.tests import ROOT_PATH
from easytext.tests import ASSERT


@ComponentRegister.register(typename="Optimizer", name_space="optimizer")
class _MyOpitmizer:
    pass


@ComponentRegister.register(typename="MyModel", name_space="model")
class _MyModel:

    def __init__(self, input_size: int, output_size: int):
        self.input_size = input_size
        self.output_size = output_size


def test_config():

    config_file_path = "data/easytext/tests/config/config.json"
    config_file_path = os.path.join(ROOT_PATH, config_file_path)

    config = Config(is_training=True,
                    config_file_path=config_file_path)

    ASSERT.assertTrue(config.model is not None)
    ASSERT.assertTrue(isinstance(config.model, _MyModel))

    ASSERT.assertTrue(config.optimizer is not None)
    ASSERT.assertTrue(isinstance(config.optimizer, _MyOpitmizer))



#!/usr/bin/env python 3
# -*- coding: utf-8 -*-

#
# Copyright (c) 2020 PanXu, Inc. All Rights Reserved
#
"""
测试 pretrained vocabulary

Authors: panxu(panxu@baidu.com)
Date:    2020/06/25 12:00:00
"""
import os
import pytest

from easytext.tests import ROOT_PATH
from easytext.tests import ASSERT

from easytext.data import PretrainedVocabulary
from easytext.data import GloveLoader
from easytext.data import Vocabulary


@pytest.fixture(scope="class")
def pretrained_vocabulary():
    """
    生成 预训练词汇表
    """
    pretrained_file_path = "data/easytext/tests/pretrained/word_embedding_sample.3d.txt"
    pretrained_file_path = os.path.join(ROOT_PATH, pretrained_file_path)

    glove_loader = GloveLoader(embedding_dim=3,
                               pretrained_file_path=pretrained_file_path)

    tokens = [["我"], ["美丽"]]

    vocab = Vocabulary(tokens=tokens,
                       padding=Vocabulary.PADDING,
                       unk=Vocabulary.UNK,
                       special_first=True)

    pretrained_vocab = PretrainedVocabulary(vocabulary=vocab,
                                            pretrained_word_embedding_loader=glove_loader)
    return pretrained_vocab


def test_pretrained_vocabulary(pretrained_vocabulary):
    """
    测试预训练词汇表
    """

    ASSERT.assertEqual(4, pretrained_vocabulary.size)
    ASSERT.assertEqual(4, len(pretrained_vocabulary))
    ASSERT.assertEqual(2, pretrained_vocabulary.index("我"))
    ASSERT.assertEqual(3, pretrained_vocabulary.index("美丽"))

    ASSERT.assertEqual((pretrained_vocabulary.size, 3),
                        pretrained_vocabulary.embedding_matrix.size())

    expect_embedding_dict = {"a": [1.0, 2.0, 3.0],
                             "b": [4.0, 5.0, 6.0],
                             "美丽": [7.0, 8.0, 9.0]}

    ASSERT.assertListEqual(expect_embedding_dict["美丽"],
                       pretrained_vocabulary.embedding_matrix[
                           pretrained_vocabulary.index("美丽")
                       ].tolist())

    zero_vec = [0.] * 3

    for index in [pretrained_vocabulary.index("我"),
                  pretrained_vocabulary.padding_index,
                  pretrained_vocabulary.index(pretrained_vocabulary.unk)]:
        ASSERT.assertListEqual(zero_vec, pretrained_vocabulary.embedding_matrix[index].tolist())


def test_save_and_load(pretrained_vocabulary):
    """
    测试 保存和载入
    """

    saved_dir = os.path.join(ROOT_PATH, "data/easytext/tests/pretrained_vocabulary")

    pretrained_vocabulary.save_to_file(saved_dir)

    loaded_vocab = PretrainedVocabulary.from_file(saved_dir)

    test_pretrained_vocabulary(loaded_vocab)








